/* Usage: cat file.log | ./nudelstats \#channel nickname language [nickfile] > file.html */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <regex.h>

/* Options */

/* Enable NO_CSS to import the theme into the binary.		*/
/* This will make the program more portable but less themeable.	*/

/* #define NO_CSS */

/* Config */
#define MAXUSERS 10000
#define MAXNICKLENGTH 50
#define MAXLINELENGTH 1000
#define MAXQUOTELENGTH 80
#define NBUSERS 50
#define NBUSERSTIME 10
#define NBURLS 8
#define NBWORDS 15
#define MINWORDLENGTH 6

/* nudelstats */
#define VERSION "0.5" // This should illustrate the current dev-version
#define URL "http://c7.ath.cx/nudelstats/" // URL to the projectsite

/* Counters */
#define D_SMILE     0
#define D_FROWN     1
#define D_EXCLAM    2
#define D_QUESTION  3
#define D_ME        4
#define D_MODE      6
#define D_KICK      7
#define D_KICKED    8
#define D_URL       9
#define D_JOIN      10
#define D_NICK      11
#define D_MONOLOGUE 12
#define NBCOUNTERS  13
char *counters[NBCOUNTERS]={"C_SMILE","C_FROWN","C_EXCLAM","C_QUESTION","C_ME","C_TOPIC","C_MODE","C_KICK","C_KICKED","C_URL","C_JOIN","C_NICK","C_MONOLOGUE"};

/* Languages */
#define NBLANGUAGES 3
#define NBKEYS 38
/* first key used for language name and abbreviation */
char *keys[NBLANGUAGES][NBKEYS+1][NBLANGUAGES]= 
{
  { /* English language */
    { "English",      "en" },
    { "HEADER",       "Statistics for %s by %s" },
    { "LEGEND",       "Legend" },
    { "LASTDAYS",     "Lastdays statistics" },
    { "TOPHOURS",     "Hourly statistics" },
    { "TOPUSERS",     "Most active people" },
    { "OTHERS",       "There are %d left not ranked..." },
    { "NBLINES",      "Lines" },
    { "NICK",         "Nick" },
    { "AVGLETTERS",   "Letters/lines" },
    { "HOURS",        "Hours" },
    { "QUOTE",        "Random message" },
    { "TOPUSERSTIME", "Most active people by time of day" },
    { "RANDTOPICS",   "Some topics" },
    { "CHANGEDBY",    "Changed by" },
    { "NEWTOPIC",     "New topic" },
    { "RANDURLS",     "Some URLs" },
    { "POSTEDBY",     "Posted by" },
    { "POSTEDURL",    "URL" },
    { "TOPWORDS",     "Most used words" },
    { "WORD",         "Word" },
    { "OCCURRENCES",  "Occurrences" },
    { "BIGNUMBERS",   "Some big numbers..." },
    { "NUMBERS",      "Numbers" },
    { "TIME",         "%d lines (%d days) parsed in %d seconds" },
    { "FOOTER",       "Statistics generated by" },
    { "C_SMILE",      "is often happy :)" },
    { "C_FROWN",      "is often sad :(" },
    { "C_EXCLAM",     "yells a lot !" },
    { "C_QUESTION",   "asks a lot of questions ?" },
    { "C_ME",         "likes /me command" },
    { "C_TOPIC",      "often changes the topic" },
    { "C_MODE",       "often changes the modes" },
    { "C_KICK",       "likes to /kick" },
    { "C_KICKED",     "is often kicked" },
    { "C_URL",        "posts many URLs" },
    { "C_JOIN",       "doesn't know wether to stay or quit" },
    { "C_NICK",       "often changes his nick" },
    { "C_MONOLOGUE",  "speaks a lot of monologues" }
  },
  { /* Swedish language by Peter Hellberg (zilver.hellberg@home.se) */
    { "Swedish",      "se" },
    { "HEADER",       "Statistik f&ouml;r %s av %s" },
    { "LEGEND",       "Legend" },
    { "LASTDAYS",     "G&aring;rdagens statistik" },
    { "TOPHOURS",     "Statistik per timme" },
    { "TOPUSERS",     "Mest aktiva" },
    { "OTHERS",       "Det &auml;r %d som inte har blivit rankade..." },
    { "NBLINES",      "rader" },
    { "NICK",         "nick" },
    { "AVGLETTERS",   "tecken/rad" },
    { "HOURS",        "timmar" },
    { "QUOTE",        "slumpat meddelande" },
    { "TOPUSERSTIME", "Mest aktiva per tid p&aring; dygnet" },
    { "RANDTOPICS",   "N&aring;gra topics" },
    { "CHANGEDBY",    "&auml;ndrat av" },
    { "NEWTOPIC",     "ny topic" },
    { "RANDURLS",     "N&aring;gra URLs" },
    { "POSTEDBY",     "postad av" },
    { "POSTEDURL",    "URL" },
    { "TOPWORDS",     "Mest anv&auml;nda ord" },
    { "WORD",         "ord" },
    { "OCCURRENCES",  "antal g&aring;nger" },
    { "BIGNUMBERS",   "N&aring;gra stora siffror..." },
    { "NUMBERS",      "siffror" },
    { "TIME",         "%d rader (%d dagar) genererad p&aring; %d sekunder" },
    { "FOOTER",       "Statistiken genererades av" },
    { "C_SMILE",      "&auml;r ofta glad :)" },
    { "C_FROWN",      "&auml;r ofta ledsen :(" },
    { "C_EXCLAM",     "skriker mycket !" },
    { "C_QUESTION",   "st&auml;ller en massa fr&aring;gor ?" },
    { "C_ME",         "tycker om /me" },
    { "C_TOPIC",      "byter ofta topic" },
    { "C_MODE",       "byter ofta modes" },
    { "C_KICK",       "tycker om att sparka ut folk" },
    { "C_KICKED",     "kickas ofta" },
    { "C_URL",        "klistrar in en massa URL:er" },
    { "C_JOIN",       "joinar en massa" },
    { "C_NICK",       "&auml;ndrar sitt nick ofta" },
    { "C_MONOLOGUE",  "pratar ofta med sig sj&auml;lv" }
  },
  { /* German language */
    { "Deutsch",      "de" },
    { "HEADER",       "Statistik f&uuml;r %s von %s" },
    { "LEGEND",       "Legende" },
    { "LASTDAYS",     "Statistik des Vortages" },
    { "TOPHOURS",     "St&uuml;ndliche Statistik" },
    { "TOPUSERS",     "Die aktivsten Personen" },
    { "OTHERS",       "Es bleiben noch %d uneingetragene" },
    { "NBLINES",      "Zeilen" },
    { "NICK",         "Nick" },
    { "AVGLETTERS",   "Zeichen/Zeile" },
    { "HOURS",        "Stunden" },
    { "QUOTE",        "Zuf&auml;llig ausgew&auml;hlte Zitate" },
    { "TOPUSERSTIME", "Die aktivsten Personen zur bestimmten Tageszeit" },
    { "RANDTOPICS",   "Ein paar Topics" },
    { "CHANGEDBY",    "Ge&auml;ndert von" },
    { "NEWTOPIC",     "Neues Topic" },
    { "RANDURLS",     "Einige URLs" },
    { "POSTEDBY",     "Geschrieben von" },
    { "POSTEDURL",    "URL" },
    { "TOPWORDS",     "Am h&auml;ufigsten benutzte Worte" },
    { "WORD",         "Wort" },
    { "OCCURRENCES",  "Vorkommen" },
    { "BIGNUMBERS",   "Ein paar grosse Zahlen..." },
    { "NUMBERS",      "Zahlen" },
    { "TIME",         "%d Zeilen (%d Tage) verarbeitet in %d Sekunden" },
    { "FOOTER",       "Statistik erstellt von" },
    { "C_SMILE",      "ist oft fr&ouml;hlich :)" },
    { "C_FROWN",      "ist oft traurig :(" },
    { "C_EXCLAM",     "schreit viel !" },
    { "C_QUESTION",   "stellt viele Fragen ?" },
    { "C_ME",         "mag das /me Kommando" },
    { "C_TOPIC",      "&Auml;ndert oft das Topic" },
    { "C_MODE",       "&Auml;ndert oft die Modes" },
    { "C_KICK",       "/kick 't gerne" },
    { "C_KICKED",     "wird oft ge /kick 't" },
    { "C_URL",        "schreibt viele URLs" },
    { "C_JOIN",       "weiss nicht ob er bleiben oder gehen soll" },
    { "C_NICK",       "&auml;ndert oft seinen Namen" },
    { "C_MONOLOGUE",  "monologisiert sehr viel" }
  }
};

int language;


char *L(char *key)
{
  int i;
  for (i=1;i<=NBKEYS;i++) if (strcmp(key,keys[language][i][0])==0) return(keys[language][i][1]);
  fprintf(stderr,"unknown language key: %s\n",key);
  return("");
}


/* Variables */

char *channel;
char *maintainer;

struct
{
  char nick[MAXNICKLENGTH];
  int lines;
  int letters;
  int hours[4];
  char quote[MAXQUOTELENGTH+1];
  int counters[NBCOUNTERS];
  int temp;
} users[MAXUSERS];
int nbusers=0;

struct
{
  char nick[MAXNICKLENGTH];
  char url[MAXLINELENGTH];
  char shorturl[MAXQUOTELENGTH+1];
} urls[5];
int nburls=0;

struct
{ 
  char nick[MAXNICKLENGTH];
  char topic[MAXQUOTELENGTH+1];
} topics[5];
int nbtopics=0;

struct
{
  int lines;
  int hours[4];
} lastdays[31];
int days=0;

int hours[24];
int lines=0;

struct letter
{
  int nb;
  struct letter *next[26];
} words;

struct
{
  int nb;
  char word[MAXLINELENGTH];
} topwords[NBWORDS];

#define isletter(c) (((c>='a')&&(c<='z'))||((c>='A')&&(c<='Z')))
#define lowercase(c) (((c>='A')&&(c<='Z'))?c-'A'+'a':c)
void findwords(char *message)
{
  int i,c;
  struct letter *pos,*tmp;
  for (;;)
  {
    while (!isletter(*message)) if (*message=='\0') return; else message++;
    pos=&words;
    while (isletter(*message))
    {
      c=lowercase(*message)-'a';
      if (pos->next[(int)c]==NULL)
      {
        tmp=malloc(sizeof(struct letter));
        tmp->nb=0;
        for (i=0;i<26;i++) tmp->next[i]=NULL;
        pos->next[(int)c]=tmp;
      }
      pos=pos->next[(int)c];
      message++; 
    }
    pos->nb++;
  }
  return;
}

char tempword[MAXLINELENGTH];
void bestwords(struct letter pos,int cur)
{
  int i,j;
  if ((cur>=MINWORDLENGTH)&&(pos.nb>topwords[NBWORDS-1].nb))
  {
    for (i=0;pos.nb<topwords[i].nb;i++);
    for (j=NBWORDS-1;j>i;j--)
    {
      topwords[j].nb=topwords[j-1].nb;
      strcpy(topwords[j].word,topwords[j-1].word);
    }
    topwords[i].nb=pos.nb;
    strcpy(topwords[i].word,tempword);
  }
  for (i=0;i<26;i++) if (pos.next[i]!=NULL)
  {
    tempword[cur]='a'+i;
    bestwords(*(pos.next[i]),cur+1);
  }
  tempword[cur]='\0';
}

void printhtml(char *string) /* replace < and > by &lt; and &gt; */
{
  while (*string!='\0')
  {
    switch (*string)
    {
      case '<':printf("&lt;"); break;
      case '>':printf("&gt;"); break;
      case '&':printf("&amp;"); break;
      default:printf("%c",*string); break;
    }
    string++;
  }
  return;
}

int dichotomic(char *nick)
{
  int i,j,start=0,end=nbusers-1,middle;
  while (start<=end)
  {
    middle=(start+end)/2;
    if (strcmp(nick,users[middle].nick)>0) start=middle+1; else end=middle-1;
  }
  if (strcmp(nick,users[start].nick)!=0)
  {
    nbusers++;
    if (nbusers>=MAXUSERS) { fprintf(stderr,"too many users\n"); exit(1); }
    for (i=nbusers-1;i>start;i--)
    {
      strcpy(users[i].nick,users[i-1].nick);
      users[i].lines=users[i-1].lines;
      users[i].letters=users[i-1].letters;
      for (j=0;j<4;j++) users[i].hours[j]=users[i-1].hours[j];
      strcpy(users[i].quote,users[i-1].quote);
      for (j=0;j<NBCOUNTERS;j++) users[i].counters[j]=users[i-1].counters[j];
      users[i].temp=users[i-1].temp;
    }
    strcpy(users[start].nick,nick);
    users[start].lines=0;
    users[start].letters=0;
    for (j=0;j<4;j++) users[start].hours[j]=0;
    users[start].quote[0]='\0';
    for (j=0;j<NBCOUNTERS;j++) users[start].counters[j]=0;
    users[start].temp=0;
  }
  return(start);
}

int main(int argc,char *argv[])
{
  int i,j,k;
  int max,user,temp,hour;
  int mononick,monolines;
  time_t debut;
  int totallines=0;
  int pos=0;
  char c;
  char *nick,*message;
  char line[MAXLINELENGTH];
  FILE *fic;
  regex_t preg;
  
  /*** INIT ***/
  
  if ((argc<4) || (argc>5))
  {
    fprintf(stderr,"Usage: cat file.log | ./nudelstats \\#channel nickname language [nickfile] > file.html\n\n");
    fprintf(stderr,"Version : nudelstats %s\n\n",VERSION);
    fprintf(stderr,"Supported languages :\n");
    for (i=0;i<NBLANGUAGES;i++) fprintf(stderr,"%s = %s\n",keys[i][0][0],keys[i][0][1]); 
    return(0);
  }

  channel=argv[1];
  maintainer=argv[2];
  for (i=0;i<NBLANGUAGES;i++) if (strcmp(argv[3],keys[i][0][1])==0) { language=i; break; }
  if (i==NBLANGUAGES)
  {
    fprintf(stderr,"Invalid language : %s\n",argv[3]);
    return(1);
  }

  /*** LOG ***/
  
  srand(debut=time(NULL));
  // Write a check for current language (same goes for all output)
  fprintf(stderr,"working on %s : ",channel);
  
  while (!feof(stdin))
  {
    c=getchar();
    line[pos++]=c;
    if (pos>=MAXLINELENGTH) { fprintf(stderr,"line %d is too long\n",totallines); exit(1); }
    if (c=='\n')
    {
      line[pos-1]='\0';
      totallines++;
      if (totallines%10000==0) { fprintf(stderr,"."); fflush(stderr); }
      if (strncmp("--- Day changed",line,15)==0) /* --- Day changed Wed May 01 2002 */
      {
        for (i=30;i>0;i--)
        {
          lastdays[i].lines=lastdays[i-1].lines;
          for (j=0;j<4;j++) lastdays[i].hours[j]=lastdays[i-1].hours[j];
        }
        lastdays[0].lines=0;
        for (j=0;j<4;j++) lastdays[0].hours[j]=0;
        days++;
      }
      else if (strncmp("-!- mode/",&line[6],9)==0) /* 00:00 -!- mode/#channel [...] by (Nick, Nick2, )Nick3 */
      {
        for (i=strlen(line);line[i]!=' ';i--);
        nick=&line[i+1];
        users[dichotomic(nick)].counters[D_MODE]++;
      }
      else if (strncmp("-!-",&line[6],3)==0) /* 00:00 -!- Nick something... */
      {
        for (i=10;line[i]!=' ';i++);
        line[i]='\0';
        nick=&line[10];
        message=&line[i+1];
         
        if (strncmp("was kicked from",message,15)==0) /* 00:00 -!- Nick was kicked from #channel by Nick [Reason] */
        {
          users[dichotomic(nick)].counters[D_KICKED]++;
          for (i=16;message[i]!=' ';i++);
          message=&message[i+4];
          for (i=0;message[i]!=' ';i++);
          message[i]='\0';
          users[dichotomic(message)].counters[D_KICK]++;
        }
        else if (strncmp("is now known as",message,15)==0) /* 00:00 -!- Nick is now known as Nick */
          users[dichotomic(nick)].counters[D_NICK]++;
        else if (message[0]=='[') /* 00:00 -!- Nick [user@host] something... */
        {
          for (i=0;message[i]!=']';i++);
          message=&message[i+2];
          if (strncmp("has joined",message,10)==0) /* 00:00 -!- Nick [user@host] has joined #channel */
            users[dichotomic(nick)].counters[D_JOIN]++;
          else if (strncmp("has quit",message,8)==0); /* 00:00 -!- Nick [user@host] has quit [Reason] */
          else if (strncmp("has left",message,8)==0); /* 00:00 -!- Nick [user@host] has left #channel [Reason] */
          else;
        }
      }
      else if ((line[6]=='<') || (line[7]=='*'))
      {
        line[2]='\0';
        hour=atoi(line);
        if (line[7]=='*') /* 00:00  * Nick the message */
        {
          for (i=9;line[i]!=' ';i++);
          nick=&line[9];
          message=&line[i+1];
        }
        else if (line[7]=='>') /* 00:00 <>>>?Nick<<<> the personal message */
                               /* 00:00 <>>?Nick<<> the personal message */
        {
          for (i=10;line[i]!='<';i++);
          nick=&line[10];
          if (line[9]=='>') nick++;
          message=&line[i+5];
        }
        else /* 00:00 <?Nick> the message */
        {
          for (i=8;line[i]!='>';i++);
          nick=&line[8];
          message=&line[i+2];
        }
        line[i]='\0';
        i=dichotomic(nick);
        if (line[7]=='*') users[i].counters[D_ME]++;
        if (i==mononick)
        {
          monolines++;
          if (monolines==5) users[i].counters[D_MONOLOGUE]++;
        }
        else
        {
          mononick=i;
          monolines=1;
        }
        j=strlen(message);
        users[i].lines++;
        users[i].letters+=j;
        users[i].hours[hour/6]++;
        lastdays[0].lines++;
        lastdays[0].hours[hour/6]++;
        lines++;
        hours[hour]++;
        if (message[j-1]=='?') users[i].counters[D_QUESTION]++;
        else if (message[j-1]=='!') users[i].counters[D_EXCLAM]++;
        else if ((message[j-3]==' ')&&(message[j-2]==':'))
        {
          if (message[j-1]==')') users[i].counters[D_SMILE]++;
          else if (message[j-1]=='(') users[i].counters[D_FROWN]++;
        }
        if (rand()%users[i].lines==0) strncpy(users[i].quote,message,MAXQUOTELENGTH);
        if (strncmp("http://",message,7)==0)
        {
          users[i].counters[D_URL]++;
          for (i=0;(message[i]!=' ') && (i<strlen(message));i++);
          message[i]='\0';
          nburls++;
          if ((nburls<=NBURLS) || (rand()%(nburls/NBURLS)==0))
          {
            temp=nburls<=NBURLS?nburls-1:rand()%NBURLS;
            strcpy(urls[temp].nick,nick);
            strcpy(urls[temp].url,message);
            strncpy(urls[temp].shorturl,message,MAXQUOTELENGTH);
          }
        }
        findwords(message);
      }
      pos=0;
    }
  }
  fprintf(stderr," done\n");

  bestwords(words,0);

  /*** ALIAS ***/
  
  if (argc==5)
  {
    for (i=0;i<nbusers;i++) users[i].temp=users[i].lines;
    if ((fic=fopen(argv[4],"rt"))==NULL) { fprintf(stderr,"can't open nick file\n"); exit(1); }
    while (fscanf(fic,"%s",line)==1)
    {
      user=dichotomic(line);
      fscanf(fic,"%s",line);
      if (regcomp(&preg,line,0)!=0) { fprintf(stderr,"error in nick file"); exit(1); }
      for (i=0;i<nbusers;i++) if ((i!=user) && (regexec(&preg,users[i].nick,0,0,0)==0) && (users[i].lines>=0))
      {
        if (users[i].temp>users[user].temp) /* for nick alias, keep the random quote of the most used nick */
        {
          strcpy(users[user].quote,users[i].quote);
          users[user].temp=users[i].temp;
        }
        users[user].lines+=users[i].lines;
        users[user].letters+=users[i].letters;
        for (j=0;j<4;j++) users[user].hours[j]+=users[i].hours[j];
        for (j=0;j<NBCOUNTERS;j++) users[user].counters[j]+=users[i].counters[j];
        /* "remove" old user */
        users[i].lines=-1;
        users[i].letters=-1;
        for (j=0;j<4;j++) users[i].hours[j]=-1;
        for (j=0;j<NBCOUNTERS;j++) users[i].counters[j]=-1;
      }
      regfree(&preg);
    }
    fclose(fic);
    /* "remove" the ignored nicks */
    i=dichotomic("<NULL>");
    users[i].lines=-1;
    users[i].letters=-1;
    for (j=0;j<4;j++) users[i].hours[j]=-1;
    for (j=0;j<NBCOUNTERS;j++) users[i].counters[j]=-1;
  }
  
  /*** XHTML 1.1 ***/

  /* header */
  printf("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\n\n");
  printf("<!-- Generated by nudelstats %s : %s -->\n",VERSION,URL);
  printf("<!-- Written by ziL (#nudel @ chatsociety) -->\n\n");
  printf("<html>\n\n<head>\n<title>");
  printf(L("HEADER"),channel,maintainer);
  printf("</title>\n <meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\" />\n");
  printf(" <style type=\"text/css\">\n");

	
  #ifdef NO_CSS
    
   FILE *input_file_pointer;
   char word[50]; /* no word above 50 characters */
   input_file_pointer = fopen("testlog/theme.css", "r");
  
   if (file_input_pointer == NULL) {
   perror("fopen"); exit(1); }
  
   printf("\n/* The stylesheet has been parsed into the binary. */\n");
   printf("/* Remove #define NO_CSS in nudelstats.c to use the file theme.css instead */\n\n");
  
   while( fscanf( input_file_pointer, " %s", word)!=EOF) /* loop through the file, word by word */
   {
 	 printf("%s ", word); /* print word */
 
 	 if(strrchr(word,'}')){ 
 		printf("\n"); /* insert a break */
 	 }
               
 	 if(strrchr(word,'/')){
 		 if(strrchr(word,'*')){ /* check if we got a * in the word */
                 	 printf("\n\n"); /* insert two breaks */
                  }
 	 }              
    } /* finished parsing the file */
  #else
   printf(" @import \"theme.css\";\n");
  #endif
	
	
  printf(" </style>\n");
  
  printf("</head>\n\n");
  printf("<body>\n\n<div class=\"ns_align\">\n\n<div class=\"ns_title_1\">");
  printf(L("HEADER"),channel,maintainer);
  printf("</div>\n%s<br />\n<br /><br />\n\n",ctime(&debut));

  /* legend */
  printf("<div id=\"ns_sect_legend\"><span class=\"ns_title_2\">%s</span><br /><br />\n<table class=\"ns_table_bg\">\n<tr>\n",L("LEGEND"));
  for (i=0;i<4;i++) printf("<td class=\"ns_alt_td\"><div class=\"ns_hour_%d\"></div></td><td class=\"ns_alt_td\">%s %d-%d &nbsp;</td>\n",i+1,L("HOURS"),i*6,i*6+5);
  printf("</tr>\n</table>\n</div>\n\n");
  
  /* last days */
  printf("<div id=\"ns_sect_lastdays\"><span class=\"ns_title_2\">%s</span><br /><br />\n<table class=\"ns_table_bg\">\n<tr>\n",L("LASTDAYS"));
  max=-1;
  for (i=24;i>=0;i--) if (lastdays[i].lines>max) max=lastdays[i].lines;
  for (i=24;i>=0;i--)
  {
    printf("<td class=\"ns_alt_td\" valign=\"bottom\">%d<br />",lastdays[i].lines);
    for (j=0;j<4;j++) if (lastdays[i].hours[j]!=0) printf("<div class=\"ns_hour_%d\" style=\"height: %dpx;\"></div>",j+1,150*lastdays[i].hours[j]/max);
    printf("</td>\n");
  }
  printf("</tr>\n<tr>\n");
  for (i=24;i>=0;i--)
    printf("<td valign=\"middle\">%d</td>\n",i);
  printf("</tr>\n</table>\n</div>\n\n");
  
  /* top hours */
  printf("<div id=\"ns_sect_tophours\"><span class=\"ns_title_2\">%s</span><br /><br />\n<table class=\"ns_table_bg\">\n<tr>\n",L("TOPHOURS"));
  max=-1;
  for (i=0;i<24;i++) if (hours[i]>max) max=hours[i];
  for (i=0;i<24;i++)
  {
    printf("<td class=\"ns_alt_td\" valign=\"bottom\">%.1f<br />",lines!=0?(float)100*hours[i]/lines:0);
    if (hours[i]!=0) printf("<div class=\"ns_hour_%d\" style=\"height: %dpx;\"></div>",i/6+1,150*hours[i]/max);
    printf("</td>\n");
  }
  printf("</tr>\n<tr>\n");
  for (i=0;i<24;i++)
    printf("<td>%d</td>\n",i);
  printf("</tr>\n</table>\n</div>\n\n");
  
  /* top users */
  printf("<div id=\"ns_sect_topusers\"><span class=\"ns_title_2\">%s</span><br /><br />\n",L("TOPUSERS"));
  printf("<table class=\"ns_table_bg\">\n<tr class=\"ns_title_3\"><td></td><td>%s</td><td>%s</td><td>%s</td><td colspan=\"2\">%s</td><td>%s</td></tr>\n",L("NICK"),L("NBLINES"),L("HOURS"),L("AVGLETTERS"),L("QUOTE"));
  for (i=1;i<=NBUSERS;i++)
  {
    user=-1;
    max=0;
    for (j=0;j<nbusers;j++) if (users[j].lines>max) max=users[user=j].lines;
    if (user!=-1)
    {
      printf("<tr><td align=\"right\">%d</td><td>%s</td><td>%d</td><td>",i,users[user].nick,users[user].lines);
      for (j=0;j<4;j++) if (users[user].hours[j]!=0) printf("<div class=\"ns_hour_%d\" style=\"width: %dpx; height: 5px;\"></div>",j+1,100*users[user].hours[j]/users[user].lines);
      printf("</td><td>%d</td><td><div class=\"ns_letters\" style=\"width: %dpx;\"></div></td><td>\"",users[user].lines!=0?users[user].letters/users[user].lines:0,users[user].lines!=0?users[user].letters/users[user].lines:0);
      printhtml(users[user].quote);
      printf("\"</td></tr>\n");
      users[user].lines=-1;
    }    
  }
  printf("</table>\n");
  temp=0;
  for (i=0;i<=nbusers;i++) if (users[i].lines>=0) temp++;
  if (temp>0)
  {
    printf("<br />");
    printf(L("OTHERS"),temp);
    printf("<br />\n");
  }
  printf("</div>\n\n");
  
  /* top users by time */
  printf("<div id=\"ns_sect_toputime\"><span class=\"ns_title_2\">%s</span><br /><br />\n",L("TOPUSERSTIME"));
  printf("<table class=\"ns_table_bg\">\n<tr class=\"ns_title_3\"><td></td>");
  for (i=0;i<4;i++) printf("<td colspan=\"2\">%s %d-%d</td>",L("HOURS"),i*6,i*6+5);
  printf("</tr>\n");
  for (i=1;i<=NBUSERSTIME;i++)
  {
    printf("<tr><td>%d</td>",i);
    for (j=0;j<4;j++)
    {
      user=-1;
      max=0;
      for (k=0;k<nbusers;k++) if (users[k].hours[j]>max) max=users[user=k].hours[j];
      if (user!=-1)
      {
        printf("<td>%s</td><td>%d</td>",users[user].nick,users[user].hours[j]);
        users[user].hours[j]=-1;
      }
      else printf("<td></td><td></td>");
    }
    printf("</tr>\n");
  }
  printf("</table>\n</div>\n\n");
 
  /* random urls */
  printf("<div id=\"ns_sect_randurls\"><span class=\"ns_title_2\">%s</span><br /><br />\n",L("RANDURLS"));
  printf("<table class=\"ns_table_bg\">\n<tr class=\"ns_title_3\"><td>%s</td><td>%s</td></tr>\n",L("POSTEDBY"),L("POSTEDURL"));
  for (i=nburls<NBURLS?nburls-1:NBURLS-1;i>=0;i--)
  {
    printf("<tr><td>%s</td><td><a href=\"",urls[i].nick);
    printhtml(urls[i].url);
    printf("\">");
    printhtml(urls[i].shorturl);
    printf("</a></td></tr>\n");
  }
  printf("</table>\n</div>\n\n");
  
  /* top words */
  printf("<div id=\"ns_sect_topwords\"><span class=\"ns_title_2\">%s</span><br /><br />\n",L("TOPWORDS"));
  printf("<table class=\"ns_table_bg\">\n<tr class=\"ns_title_3\"><td></td><td>%s</td><td>%s</td></tr>\n",L("WORD"),L("OCCURRENCES"));
  for (i=0;i<NBWORDS;i++)
    if (topwords[i].nb!=0) printf("<tr><td align=\"right\">%d</td><td>\"%s\"</td><td>%d</td></tr>\n",i+1,topwords[i].word,topwords[i].nb);
  printf("</table>\n</div>\n\n");

  /* big numbers */
  printf("<div id=\"ns_sect_bignums\"><span class=\"ns_title_2\" >%s</span><br /><br />\n",L("BIGNUMBERS"));
  printf("<table class=\"ns_table_bg\">\n<tr class=\"ns_title_3\"><td>%s</td><td>%s</td><td>%s</td></tr>\n",L("NICK"),L("NUMBERS"),L("NBLINES"));
  for (i=0;i<NBCOUNTERS;i++)
  {
    user=-1;
    max=0;
    for (j=0;j<nbusers;j++) if (users[j].counters[i]>max) max=users[user=j].counters[i];
    if (user!=-1) printf("<tr><td>%s</td><td>%s</td><td>%d</td></tr>",users[user].nick,L(counters[i]),users[user].counters[i]);
  }
  printf("</table>\n</div>\n\n");
  
  /* footer */
  printf(L("TIME"),totallines,days,(int)(time(NULL)-debut));
  printf("<br />\n%s <a href=\"%s\">nudelstats %s</a>",L("FOOTER"),URL,VERSION);
  printf("<br /><div id=\"ns_sect_w3image\"><br />\n\n<img src=\"http://www.w3.org/Icons/valid-xhtml11\" id=\"ns_w3c_image\" height=\"31\" width=\"88\" alt=\"Probably Valid XHTML 1.1!\" title=\"Probably Valid XHTML 1.1!\" onclick=\"javascript:window.location.href='http://validator.w3.org/check/referer'\" />");
  printf("\n</div>\n</div>\n\n</body>\n\n</html>\n");
  
  return(0);
}
